# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['falocalrepo', 'falocalrepo.console']

package_data = \
{'': ['*']}

install_requires = \
['click-help-colors>=0.9.1,<0.10.0',
 'click>=8.0.3,<9.0.0',
 'faapi>=3.3.6,<3.4.0',
 'falocalrepo-database>=5.0.7,<5.1.0',
 'falocalrepo-server>=2.0.2,<2.1.0',
 'pretty-errors>=1.2.25,<2.0.0',
 'psutil>=5.9.0,<6.0.0',
 'wcwidth>=0.2.5,<0.3.0']

entry_points = \
{'console_scripts': ['falocalrepo = falocalrepo.__main__:main']}

setup_kwargs = {
    'name': 'falocalrepo',
    'version': '4.0.2',
    'description': "Pure Python program to download any user's gallery, scraps, favorites, and journals from FurAffinity in an easily handled database.",
    'long_description': '<div align="center">\n\n<img alt="logo" width="400" src="https://gitlab.com/uploads/-/system/project/avatar/6750403/falocalrepo.png">\n\n# FALocalRepo\n\nPure Python program to download submissions, journals, and user folders from the FurAffinity forum in an easily handled\ndatabase.\n\n[![version_pypi](https://img.shields.io/pypi/v/falocalrepo?logo=pypi)](https://pypi.org/project/falocalrepo/)\n[![version_python](https://img.shields.io/pypi/pyversions/falocalrepo?logo=Python)](https://www.python.org)\n\n[![version_gitlab](https://img.shields.io/gitlab/v/tag/6750403?label=gitlab&sort=date&logo=gitlab&color=FCA121)](https://gitlab.com/projects/6750403)\n[![issues_gitlab](https://img.shields.io/badge/dynamic/json?logo=gitlab&color=orange&label=issues&suffix=%20open&query=%24.length&url=https%3A%2F%2Fgitlab.com%2Fapi%2Fv4%2Fprojects%2F6750403%2Fissues%3Fstate%3Dopened)](https://gitlab.com/MatteoCampinoti94/FALocalRepo/issues)\n\n[![furaffinity](https://furry-badges.herokuapp.com/badge/user/FurAffinity/FlameOfFurious)](https://furaffinity.net/user/FlameOfFurious)\n\n</div>\n\n## Introduction\n\nThis program was born with the desire to provide a relatively easy-to-use method for FA users to download submissions\nthat they care about from the forum.\n\nThe data is stored into a SQLite database, and the submissions files are saved in a tiered tree structure based on their\nID\'s. Using SQLite instead of a client-server database makes the program extremely portable, only needing a working\nPython 3.10+ installation to work, and allows the downloaded data to be moved and backed up by simply moving/copying the\ndatabase file and submissions files folder.\n\nAll download operations are performed through the custom FurAffinity scraping\nlibrary [faapi](https://pypi.org/project/faapi/). To ensure proper crawling behavior the library strictly follows\nFurAffinity\'s [robots.txt](https://www.furaffinity.net/robots.txt) in regard to allowed paths and crawl delay.\nFurthermore, submission files downloads are throttled to 100 KB/s to ensure the program won\'t use too much bandwidth.\n\nThe database and file-storage functions are handled independently by\nthe [falocalrepo-database](https://pypi.org/project/falocalrepo-database/) package which performs all transactions,\nqueries, and file operations.\n\nThe [falocalrepo-server](https://pypi.org/project/falocalrepo-server/) package is used to provide the server\nfunctionalities of the program.\n\n## Contents\n\n1. [Installation and Update](#installation-and-update)\n1. [Cookies](#cookies)\n1. [Usage](#usage)\n    1. [Environmental Variables](#environmental-variables)\n    1. [Init](#init)\n    1. [Help](#help)\n    1. [Config](#config)\n    1. [Download](#download)\n    1. [Database](#database)\n        1. [Database Query Language](#database-query-language)\n    1. [Server](#server)\n    1. [Completions](#completions)\n    1. [Updates](#updates)\n1. [Database](#database-1)\n    1. [Settings](#settings)\n    1. [Users](#users)\n    1. [Submissions](#submissions)\n    1. [Journals](#journals)\n1. [Submission Files](#submission-files)\n1. [Upgrading Database](#upgrading-database)\n1. [Contributing](#contributing)\n1. [Issues](#issues)\n1. [Appendix](#appendix)\n\n## Installation and Update\n\nTo install the program it is sufficient to use Python pip and get the package `falocalrepo`.\n\n```shell\npython3 -m pip install falocalrepo\n```\n\nPython 3.10 or above is needed to run this program, all other dependencies are handled by pip during installation. For\ninformation on how to install Python on your computer, refer to the official\nwebsite [Python.org](https://www.python.org/).\n\nTo upgrade the `falocalrepo` and its dependencies, use pip to upgrade all three components.\n\n```shell\npython3 -m pip install --upgrade falocalrepo faapi falocalrepo-database falocalrepo-server\n```\n\nTo check for updates use the [`updates` command](#updates). A message will appear if there is an update available for\nany component.\n\nThe program needs cookies from a logged-in FurAffinity session to download protected pages. Without the cookies the\nprogram can still download publicly available pages, but others will return empty. See [#Cookies](#cookies) for more\ndetails on which cookies to use.\n\n**Warning**: FurAffinity theme template must be set to "modern". Can be changed\nat [furaffinity.net/controls/settings/](https://www.furaffinity.net/controls/settings/).\n\n## Cookies\n\nThe scraping library used by this program needs two specific cookies from a logged-in FurAffinity session. These are\ncookie `a` and cookie `b`. The cookies\' values usually take the form of hexadecimal strings\nlike `356f5962-5a60-0922-1c11-65003b703038`.\n\nThe easiest way to obtain these cookies is by using a browser extension to extract them and then search for `a`\nand `b`.<br>\nAlternatively, the storage inspection tool of a desktop browser can also be used. For example on Mozilla\'s Firefox this\ncan be opened with &#8679;F9, on Safari with &#8997;&#8984;I, etc.\n\nTo set the cookies use the `config cookies` command. See [#Configuration](#config) for more details.\n\n### Environmental Variables\n\n`falocalrepo` supports the following environmental variables:\n\n* `FALOCALREPO_DATABASE` sets a path for the database rather than using the current folder.\n* `FALOCALREPO_MULTI_CONNECTION` allow operating on the database even if it is already opened in other processes.<br/>\n  **Warning**: using this option may cause the database to become corrupt and irreparable.\n* `FALOCALREPO_NO_COLOR` turn off colors for all commands.\n\n## Usage\n\n> **How to Read Usage Instructions**\n>  * `command` a static command keyword\n>  * `{arg1|arg2}` mutually exclusive arguments, only use one\n>  * `<arg>` `<param>` `<value>` an argument, parameter, value, etc. that must be provided to a command\n>  * `--option <value>` an option argument with a static key and a value\n>  * `[<arg>]` an optional argument that can be omitted\n\nTo run the program, simply call `falocalrepo` in your shell after installation.\n\nRunning without arguments will prompt a help message with all the available options and commands.\n\nThe usage pattern for the program is as follows:\n\n```\nfalocalrepo [OPTIONS] COMMAND [ARGS]...\n```\n\nAvailable options are:\n\n* `--version` Show version and exit.\n* `--versions` Show components\' versions and exit.\n\nThe following global options are available for all commands:\n\n* `--database` Specify a database different from the default (FA.db in local folder). Overrides `FALOCALREPO_DATABASE`\n  environment variable.\n* `--color / --no-color` Toggle ANSI colors.\n* `-h, --help` Show help message and exit.\n\nAvailable commands are:\n\n* `init` Initialise the database.\n* `help` Show the help for a command.\n* `config` Change settings.\n* `download` Download resources.\n* `database` Operate on the database.\n* `server` Start local server to browse database.\n* `completions` Generate tab-completion scripts.\n* `updates` Check for updates to components.\n\n_Note:_ only one connection to a database is allowed at any given time, if the database is opened in other processes,\nthe program will close with an error.\n\n_Note_: the program will not operate if the version of the database does not match the version of\nthe `falocalrepo-database` module. Only `database info` and `database upgrade` commands can be run if the database is\nnot up-to-date.\n\nWhen the database is first initialised, it sets the submissions files folder to `FA.files` (relative to the database\nlocation). This value can be changed using the [`config` command](#config).\n\nCookies need to be set manually with the config command before the program will be able to access protected pages.\n\n### Init\n\n`init`\n\nThe `init` command initialises the database. If a database is already present, no operation is performed except for a\nversion check.\n\n### Config\n\n`config <OPERATION>`\n\nThe config command allows reading and changing the settings used by the program.\n\nAvailable config commands are:\n\nAvailable settings are:\n\n* `list` List settings.\n* `cookies [--cookie <NAME> <VALUE>...]` Read or modify stored cookies. If no `--cookie` option is given, the current\n  values are read instead.\n\n> ```\n> falocalrepo config cookies --cookie a 38565475-3421-3f21-7f63-3d341339737 --cookie b 356f5962-5a60-0922-1c11-65003b703038\n> ```\n\n* `files-folder [--move | --no-move] [--relative | --absolute] [<NEW_FOLDER>]` Read or modify the folder used to store\n  submission files, where `NEW_FOLDER` is the path to the new folder. If `NEW_FOLDER` is omitted, the current value is\n  read instead.<br/>\n  By default, `NEW_FOLDER` is considered to be relative to the database folder. Absolute values are allowed as long as a\n  relative path to the database parent folder can exist. To force the use of an absolute value, activate the\n  `--absolute` option.\n\n> ```\n> falocalrepo config files-folder --no-move FA.files2\n> ```\n\n### Download\n\n`download <OPERATION>`\n\nThe download command performs all download operations to save and update users, submissions, and journals. Submissions\nare downloaded together with their files and thumbnails, if there are any. For details on submission files,\nsee [Submission Files](#submission-files).\n\nAll download operations (except login) support the `--dry-run` option. When this is active, the database is not\nmodified, nor are submission files downloaded. Entries are simply listed and the program checks whether they are in the\ndatabase or not.\n\nAvailable operations are:\n\n* `login` Check whether the cookies stored in the database belong to a login Fur Affinity session.\n\n* `users [--dry-run] -u <USER>... -f <FOLDER>...` Download specific user folders, where `FOLDER` is one of gallery,\n  scraps, favorites, journals, userpage. Multiple `--user` and `--folder` arguments can be passed.\n\n> ```\n> falocalrepo download users -u tom -u jerry -f gallery -f scraps -f journals\n> ```\n> ```\n> falocalrepo download users tom,jerry list-favorites\n> ```\n\n* `update [--dry-run] [--deactivated] [--stop N] [-u <USER>...] [-f <FOLDER>...]` Download new entries using the users\n  and folders already in the database. `--user` and `--folder` options can be used to restrict the update to specific\n  users and or folders, where `FOLDER` is one of gallery, scraps, favorites, journals, userpage. Multiple `--user`\n  and `--folder` arguments can be passed.<br/>\n  If the `--deactivated` option is used, deactivated users are fetched instead of ignore. If the user is no longer\n  inactive, the database entry will be modified as well.<br/>\n  The `--stop` option allows setting how many entries of each folder should be found in the database before stopping the\n  update.\n\n> ```\n> falocalrepo download update --stop 5\n> ```\n> ```\n> falocalrepo download update --deactivated -f gallery -f scraps\n> ```\n> ```\n> falocalrepo download update -u tom -u jerry -f favorites\n> ```\n\n* `submissions [--replace] <SUBMISSION_ID>...` Download single submissions, where `SUBMISSION_ID` is the ID of the\n  submission. If the `--replace` option is used, database entries will be overwritten with new data (favorites will be\n  maintained).\n\n> ```\n> falocalrepo download submissions 12345678 13572468 87651234\n> ```\n\n* `journals [--replace] <JOURNAL_ID>...` Download single journals, where `JOURNAL_ID` is the ID of the journal.<br/>\n  If the `--replace` option is used, database entries will be overwritten with new data (favorites will be maintained).\n\n> ```\n> falocalrepo download journals 123456 135724 876512\n> ```\n\n### Database\n\n`database <OPERATION>`\n\nOperate on the database to add, remove, or search entries. For details on columns see [Database](#database-1).\n\nAvailable operations are:\n\n* `info` show database information, statistics and version.\n* `history [--filter FILTER] [--clear]` Show database history. History events can be filtered using the `--filter`\n  option to match events that contain `FILTER` (the match is performed case-insensitively).<br/>\n  Using the `--clear` option will delete all history entries, or the ones containing `FILTER` if the `--filter` option\n  is used.\n\n> ```\n> falocalrepo database history --filter upgrade \n> ```\n\n* `search [--column <COLUMN[,WIDTH]>...] [--limit LIMIT] [--offset OFFSET] [--sort COLUMN] [--order {asc|desc}] [--output {table|csv|tsv|json|none}] [--ignore-width] [--sql] [--show-sql] [--total] {SUBMISSIONS|JOURNALS|USERS} <QUERY>...`\n  Search the database using queries, and output in different formats. For details on the query language,\n  see [Database Query Language](#database-query-language).<br/>\n  The default output format is a table with only the most relevant columns displayed for each entry. To override the\n  displayed column, or change their width, use the --column option to select which columns will be displayed (SQLite\n  statements are supported). The optional `WIDTH` value can be added to format that specific column when the output is\n  set to table.<br/>\n  To output all columns and entries of a table, `COLUMN` and `QUERY` values can be set to `@` and `%` respectively.\n  However, the\n  `database export` command is better suited for this task.<br/>\n  Search is performed case-insensitively.<br/>\n  The output can be set to five different types:\n    * `table` Table format\n    * `csv` CSV format (comma separated)\n    * `tsv` TSV format (tab separated)\n    * `json` JSON format\n    * `none` Do not print results to screen\n\n> ```\n> falocalrepo search USERS --output json \'@folders ^gallery\'\n> ```\n\n> ```\n> falocalrepo search SUBMISSIONS \'@tags |cat| |mouse| @date 2020- @category artwork\' --sort AUTHOR\n> ```\n\n> ```\n> falocalrepo search JOURNALS --putput csv \'@date (2020- | 2019-) @content commission\'\n> ```\n\n* `export [--column <COLUMN>...] [--sort COLUMN] [--order {asc|desc}] [--total] {SUBMISSIONS|JOURNALS|USERS} {csv|tsv|json} [FILE]`\n  Export all entries in a table to a file. The `FILE` argument can be omitted to print the results directly in the\n  terminal. The results total is not printed to file if a file is used.<br/>\n  By default, all columns of the table are selected, but this can be overridden with the `--column` option (SQLite\n  statements are supported).<br/>\n  Only sort and order statements are supported for exporting, to filter results use the `database search` command.<br/>\n  The `OUTPUT` can be set to four different types:\n    * `csv` CSV format (comma separated)\n    * `tsv` TSV format (tab separated)\n    * `json` JSON format\n\n> ```\n> falocalrepo export USERS --output json users.json\'\n> ```\n\n> ```\n> falocalrepo export SUBMISSIONS --sort AUTHOR\n> ```\n\n> ```\n> falocalrepo export JOURNALS --putput csv \'@date (2020- | 2019-) @content commission\'\n> ```\n\n* `add [--replace] [--submission-file FILENAME] [--submission-thumbnail FILENAME] {SUBMISSIONS|JOURNALS|USERS} <FILE>`\n  Add entries and submission files manually using a JSON file. Submission files/thumbnails can be added using the\n  respective options.<br/>\n  The JSON file must contain fields for all columns of the table. For a list of columns for each table,\n  see [Database](#database-1). By default, the program will throw an error when trying to add an entry that already\n  exists. To override this behaviour and ignore existing entries, use the `--replace` option.\n\n> ```\n> falocalrepo database add USERS user.json\n> ```\n\n* `remove [--yes] {SUBMISSIONS|JOURNALS|USERS} ID...` Remove entries from the database using their IDs. The program will\n  prompt for a confirmation before commencing deletion. To confirm deletion ahead, use the `--yes` option.\n\n> ```\n> falocalrepo database remove SUBMISSIONS 1 5 14789324\n> ```\n\n* `merge [--query <TABLE QUERY>...] [--replace] DATABASE_ORIGIN` Merge database from `DATABASE_ORIGIN`.<br/>\n  Specific tables can be selected with the `--query` option. For details on the syntax for the `QUERY` value, see\n  [Database Query Language](#database-query-language). To select all entries in a table, use `%` as query. The `TABLE`\n  value can be one of SUBMISSIONS, JOURNALS, USERS.<br/>\n  If no `--query` option is given, all major tables from the origin database are copied (SUBMISSIONS, JOURNALS, USERS).\n\n> ```\n> falocalrepo database merge ~/FA.db --query USERS tom --SUBMISSIONS \'@author tom\'\n> ```\n\n* `copy [--query <TABLE QUERY>...] [--replace] DATABASE_DEST` Copy database to `DATABASE_DEST`.<br/>\n  Specific tables can be selected with the `--query` option. For details on the syntax for the `QUERY` value, see\n  [Database Query Language](#database-query-language). To select all entries in a table, use `%` as query. The `TABLE`\n  value can be one of SUBMISSIONS, JOURNALS, USERS.<br/>\n  If no `--query` option is given, all major tables from the origin database are copied (SUBMISSIONS, JOURNALS, USERS).\n\n> ```\n> falocalrepo database copy ~/FA.db --query USERS tom --SUBMISSIONS \'@author tom\'\n> ```\n\n* `clean` Clean the database using the SQLite [VACUUM](https://www.sqlite.org/lang_vacuum.html) function.\n* `upgrade` Upgrade the database to the latest version.\n\n#### Database Query Language\n\nThe query language used for search queries is based and improves upon the search syntax currently used by the Fur\nAffinity website. Its basic elements are:\n\n* `@<field>` field specifier (e.g. `@title`), all database columns are available as search fields.\n* `()` parentheses, they can be used for better logic operations\n* `&` _AND_ logic operator, used between search terms\n* `|` _OR_ logic operator, used between search terms\n* `!` _NOT_ logic operator, used as prefix of search terms\n* `""` quotes, allow searching for literal strings without needing to escape\n* `%` match 0 or more characters\n* `_` match exactly 1 character\n* `^` start of field, when used at the start of a search term it matches the beginning of the field\n* `$` end of field, when used at the end of a search term it matches the end of the field\n\nAll other strings are considered search terms.\n\nThe search uses the `@any` field by default, allowing to do general searches without specifying a field.\n\nSearch terms that are not separated by a logic operator are considered AND terms (i.e. `a b c` &harr; `a & b & c`).\n\nExcept for the `ID`, `AUTHOR`, and `USERNAME` fields, all search terms are matched by fields containing the term: i.e.\n`@description cat` will match any item whose description field contains "cat". To match items that contain only "cat" (\nor start with, end with, etc.), the `%`, `_`, `^`, and `$` operators need to be used (e.g. `@description ^cat`).\n\nSearch terms for `ID`, `AUTHOR`, and `USERNAME` are matched exactly as they are: i.e. `@author tom` will match only\nitems whose author field is exactly equal to "tom", to match items that contain "tom" the `%`, `_`, `^`, and `$`\noperators need to be used (e.g. `@author %tom%`).\n\n### Server\n\n`server [--host HOST] [--port PORT] [--ssl-cert FILE] [--ssl-key FILE] [--redirect-http PORT2] [--auth USERNAME:PASSWORD] [--precache]`\n\nStart a server at `HOST`:`PORT` to navigate the database. The `--ssl-cert` and `--ssl-cert` allow serving with HTTPS.\nUsing `--redirect-http` starts the server in HTTP to HTTPS redirection mode. `--auth` enables HTTP Basic\nauthentication. `--precache` caches database entries at startup. For more details on usage see\n[falocalrepo-server](https://pypi.org/project/falocalrepo_server/).\n\n### Completions\n\n`completions [--alias NAME] {bash|fish|zsh}`\n\nGenerate tab-completion scripts for your shell. The generated completion must be saved in the correct location for it to\nbe recognized and used by the shell. The optional `--alias` option allows generating completion script with a name other\nthan `falocalrepo`.\n\nSupported shells are:\n\n* `bash` The Bourne Again SHell\n* `fish` The friendly interactive shell\n* `zsh` The Z shell\n\n### Updates\n\n`updates [--shell]`\n\nCheck for updates to falocalrepo and its main dependencies on PyPi. The optional `--shell` option can be used to output\nthe shell command to upgrade any component that has available updates.\n\n_Note_: The command needs an internet connection.\n\n### Help\n\n`help [<COMMAND>...]`\n\nThe `help` command gives information on the usage of the program and its commands.\n\n> ```\n> falocalrepo help database search\n> ```\n\n## Database\n\nTo store the metadata of the downloaded submissions, journals, users, cookies and statistics, the program uses a SQLite3\ndatabase. This database is built to be as light as possible while also containing all the metadata that can be extracted\nfrom a submission page.\n\nTo store all this information, the database uses four tables: `SETTINGS`, `USERS`, `SUBMISSIONS` and `JOURNALS`.\n\n> **How Lists Are Stored**<br>\n> Some fields in the database table contain lists of items. These are stored as strings, with each item surrounded by\n> bars (`|`). This allows to properly separate and search individual items regardless of their position in the list.<br>\n> `|item1||item2|`<br>\n\n### Users\n\nThe users table contains a list of all the users that have been download with the program, the folders that have been\ndownloaded, and the submissions found in each of those.\n\nEach entry contains the following fields:\n\n* `USERNAME` The URL username of the user (no underscores or spaces)\n* `FOLDERS` the folders downloaded for that specific user, sorted and bar-separated\n* `USERPAGE` the user\'s profile text\n\n### Submissions\n\nThe submissions table contains the metadata of the submissions downloaded by the program and information on their files\n\n* `ID` the id of the submission\n* `AUTHOR` the username of the author (uploader) in full format\n* `TITLE`\n* `DATE` upload date in ISO format _YYYY-MM-DDTHH:MM_\n* `DESCRIPTION` description in html format\n* `TAGS` bar-separated tags\n* `CATEGORY`\n* `SPECIES`\n* `GENDER`\n* `RATING`\n* `TYPE` image, text, music, or flash\n* `FILEURL` the remote URL of the submission file\n* `FILEEXT` the extensions of the downloaded file. Can be empty if the file contained errors and could not be recognised\n  upon download\n* `FILESAVED` file and thumbnail download status as a 2bit flag: `1x` if the file was downloaded `0x` if not, `x1` if\n  thumbnail was downloaded, `x0` if not. Possible values are `0`, `1`, `2`, `3`.\n* `FAVORITE` a bar-separated list of users that have "faved" the submission\n* `MENTIONS` a bar-separated list of users that are mentioned in the submission description as links\n* `FOLDER` the folder of the submission (`gallery` or `scraps`)\n* `USERUPDATE` whether the submission was added as a user update or favorite/single entry\n\n### Journals\n\nThe journals table contains the metadata of the journals downloaded by the program.\n\n* `ID` the id of the journal\n* `AUTHOR` the username of the author (uploader) in full format\n* `TITLE`\n* `DATE` upload date in ISO format _YYYY-MM-DDTHH:MM_\n* `CONTENT` content in html format\n* `MENTIONS` a bar-separated list of users that are mentioned in the journal content as links\n* `USERUPDATE` whether the journal was added as a user update or single entry\n\n### Settings\n\nThe settings table contains settings for the program and variable used by the database handler and main program.\n\n* `COOKIES` cookies for the scraper, stored in JSON format\n* `FILESFOLDER` location of downloaded submission files\n* `VERSION` database version, this can differ from the program version\n\n### History\n\nThe history table holds events related to the database.\n\n* `TIME` event time in ISO format _YYYY-MM-DDTHH:MM:SS.ssssss_\n* `EVENT` the event description\n\n## Submission Files\n\nSubmission files are saved in a tiered tree structure based on their submission ID. IDs are zero-padded to 10 digits and\nthen broken up in 5 segments of 2 digits; each of these segments represents a folder tha will be created in the tree.\n\nFor example, a submission `1457893` will be padded to `0001457893` and divided into `00`, `01`, `45`, `78`, `93`. The\nsubmission file will then be saved as `00/01/45/78/93/submission.file` with the correct extension extracted from the\nfile itself - FurAffinity links do not always contain the right extension.\n\n## Upgrading Database\n\nWhen the program starts, it checks the version of the database against the one used by the program and if the latter is\nmore advanced it upgrades the database.\n\n_Note:_ versions prior to 4.19.0 are not supported by falocalrepo-database version 5.0.0 and above. To update from\nthose, use [falocalrepo version 3.25.0](https://pypi.org/project/falocalrepo/v3.25.0) to upgrade the database to version\n4.19.0 first.\n_Note:_ Versions before 2.7.0 are not supported by falocalrepo version 3.0.0 and above. To update from those to the new\nversion use version 2.11.2 to update the database to version 2.7.0\n\nFor details on upgrades and changes between database versions,\nsee [falocalrepo-database](https://pypi.org/project/falocalrepo-database/).\n\n## Contributing\n\nAll contributions and suggestions are welcome!\n\nThe only requirement is that any merge request must be sent to the GitLab project as the one on GitHub is only a\nmirror: [GitLab/FALocalRepo](https://gitlab.com/MatteoCampinoti94/FALocalRepo)\n\nIf you have suggestions for fixes or improvements, you can open an issue with your idea, see [#Issues](#issues) for\ndetails.\n\n## Issues\n\nIf you encounter any problem while using the program, an issue can be opened on the project\'s pages\non [GitLab](https://gitlab.com/MatteoCampinoti94/FALocalRepo/issues).\n\nIssues can also be used to suggest improvements and features.\n\nWhen opening an issue for a problem, please copy the error message and describe the operation in progress when the error\noccurred.\n\n## Appendix\n\n### Earlier Releases\n\nRelease 3.0.0 was deleted from PyPi because of an error in the package information. However, it can still be found in\nthe code repository under tag [v3.0.0](https://gitlab.com/MatteoCampinoti94/FALocalRepo/tags/v3.0.0).\n\nRelease binaries for versions 2.11.x can be found on GitLab\nat [FALocalRepo/tags 2.11](https://gitlab.com/MatteoCampinoti94/FALocalRepo/tags?search=v2.11).\n',
    'author': 'Matteo Campinoti',
    'author_email': 'matteo.campinoti94@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://gitlab.com/MatteoCampinoti94/FALocalRepo',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)
